﻿namespace HeuristicLab.Algorithms.PushGP.Expressions
{
    using System;
    using HeuristicLab.Algorithms.PushGP.Interpreter;

    /// <summary>
    ///     Pushes the sum of the top two items.
    /// </summary>
    public class IntegerAddExpression : PushResultExpression<long>
    {
        public IntegerAddExpression() : base("INTEGER.+")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            this.Eval(interpreter.IntegerStack, 2, values => values[0] + values[1]);
        }
    }

    /// <summary>
    ///     Pushes the difference of the top two items; that is, the second item minus the top item.
    /// </summary>
    public class IntegerSubtractExpression : PushResultExpression<long>
    {
        public IntegerSubtractExpression() : base("INTEGER.-")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            this.Eval(interpreter.IntegerStack, 2, values => values[0] - values[1]);
        }
    }

    /// <summary>
    ///     Pushes the product of the top two items.
    /// </summary>
    public class IntegerMultiplyExpression : PushResultExpression<long>
    {
        public IntegerMultiplyExpression() : base("INTEGER.*")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            this.Eval(interpreter.IntegerStack, 2, values => values[0] * values[1]);
        }
    }

    /// <summary>
    ///     Pushes the quotient of the top two items; that is, the second item divided by the top item.
    ///     If the top item is zero this acts as a NOOP.
    /// </summary>
    public class IntegerDivideExpression : PushResultExpression<long>
    {
        public IntegerDivideExpression() : base("INTEGER./")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            this.Eval(interpreter.IntegerStack, 2, values => values[0] / values[1], 0);
        }
    }

    /// <summary>
    ///     Pushes the second stack item modulo the top stack item. If the top item is zero this acts as a NOOP. The modulus is
    ///     computed as the
    ///     remainder of the quotient, where the quotient has first been truncated toward negative infinity. (This is taken
    ///     from the definition
    ///     for the generic MOD function in Common Lisp, which is described for example at
    ///     http://www.lispworks.com/reference/HyperSpec/Body/f_mod_r.htm.)
    /// </summary>
    public class IntegerModuloExpression : PushResultExpression<long>
    {
        public IntegerModuloExpression() : base("INTEGER.%")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            this.Eval(interpreter.IntegerStack, 2, values => values[0] % values[1], 0);
        }
    }

    /// <summary>
    ///     Pushes the minimum of the top two items.
    /// </summary>
    public class IntegerMinExpression : PushResultExpression<long>
    {
        public IntegerMinExpression() : base("INTEGER.MIN")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            this.Eval(interpreter.IntegerStack, 2, values => Math.Min(values[0], values[1]));
        }
    }

    /// <summary>
    ///     Pushes the maximum of the top two items.
    /// </summary>
    public class IntegerMaxExpression : PushResultExpression<long>
    {
        public IntegerMaxExpression() : base("INTEGER.MAX")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            this.Eval(interpreter.IntegerStack, 2, values => Math.Max(values[0], values[1]));
        }
    }

    /// <summary>
    ///     Pushes TRUE onto the BOOLEAN stack if the second item is less than the top item, or FALSE otherwise.
    /// </summary>
    public class IntegerSmallerThanExpression : PushResultExpression<long>
    {
        public IntegerSmallerThanExpression() : base("INTEGER.<")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            this.Eval(interpreter.IntegerStack, interpreter.BooleanStack, 2, values => values[0] < values[1]);
        }
    }

    /// <summary>
    ///     Pushes TRUE onto the BOOLEAN stack if the second item is greater than the top item, or FALSE otherwise.
    /// </summary>
    public class IntegerGreaterThanExpression : PushResultExpression<long>
    {
        public IntegerGreaterThanExpression() : base("INTEGER.>")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            this.Eval(interpreter.IntegerStack, interpreter.BooleanStack, 2, values => values[0] > values[1]);
        }
    }

    /// <summary>
    ///     Pushes 1 if the top BOOLEAN is TRUE, or 0 if the top BOOLEAN is FALSE.
    /// </summary>
    public class IntegerFromBooleanExpression : StatelessExpression
    {
        public IntegerFromBooleanExpression() : base("INTEGER.FROMBOOLEAN")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            if (interpreter.BooleanStack.Count == 0) return;

            var condition = interpreter.BooleanStack.Pop();
            var value = condition ? 1 : 0;

            interpreter.IntegerStack.Push(value);
        }
    }

    /// <summary>
    ///     Pushes the result of truncating the top FLOAT.
    /// </summary>
    public class IntegerFromFloatExpression : StatelessExpression
    {
        public IntegerFromFloatExpression() : base("INTEGER.FROMFLOAT")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            if (interpreter.FloatStack.Count == 0) return;

            var value = (int) interpreter.FloatStack.Pop();

            interpreter.IntegerStack.Push(value);
        }
    }
}